(* Author:  Christian Sternagel *)

theory Week08
imports Main
begin

declare [[names_short]]

section \<open>Syntax of Propositional Formulas\<close>

text \<open>
  Propositional formulas (consisting of bottom "\<bottom>", atoms,
  negation, and implication) as algebraic datatype.

  \<^bold>N\<^bold>o\<^bold>t\<^bold>e\<^bold>s\<^bold>:
  \<bullet> parentheses are for readability only
  \<bullet> double quotes are mandatory
\<close>

datatype 'a form =
  Bot
| Atom 'a
| Neg "('a form)"
| Conj "('a form)" "('a form)"
| Disj "('a form)" "('a form)"
| Impl "('a form)" "('a form)"


section \<open>The Meaning of Propositional Formulas\<close>

text \<open>
  Evaluating formulas w.r.t. a valuation (i.e.,
  function from propositional atoms to booleans).
\<close>

fun eval :: "('a \<Rightarrow> bool) \<Rightarrow> 'a form \<Rightarrow> bool"
where
  "eval v Bot \<longleftrightarrow> False"
| "eval v (Atom x) \<longleftrightarrow> v x"
| "eval v (Neg \<phi>) \<longleftrightarrow> \<not> eval v \<phi>"
| "eval v (Conj \<phi> \<psi>) \<longleftrightarrow> eval v \<phi> \<and> eval v \<psi>"
| "eval v (Disj \<phi> \<psi>) \<longleftrightarrow> eval v \<phi> \<or> eval v \<psi>"
| "eval v (Impl \<phi> \<psi>) \<longleftrightarrow> \<not> eval v \<phi> \<or> eval v \<psi>"

text \<open>
  The meaning of a formula only depends on propositional atoms
  actually occurring in it (sometimes called "Coincidence Lemma"
  or also "Extensionality Lemma").
\<close>

fun atoms :: "'a form \<Rightarrow> 'a set"
where
  "atoms Bot = {}"
| "atoms (Atom x) = {x}"
| "atoms (Neg \<phi>) = atoms \<phi>"
| "atoms (Conj \<phi> \<psi>) = atoms \<phi> \<union> atoms \<psi>"
| "atoms (Disj \<phi> \<psi>) = atoms \<phi> \<union> atoms \<psi>"
| "atoms (Impl \<phi> \<psi>) = atoms \<phi> \<union> atoms \<psi>"

lemma atoms_eval:
  "(\<forall>x \<in> atoms \<phi>. v x = w x) \<longrightarrow> eval v \<phi> = eval w \<phi>"
by (induct \<phi>) auto


subsection \<open>Satisfiability and Validity\<close>

definition "sat \<phi> \<longleftrightarrow> (\<exists>v. eval v \<phi>)"
definition "valid \<phi> \<longleftrightarrow> (\<forall>v. eval v \<phi>)"


section \<open>Conjunctive Normal Forms\<close>

text \<open>Literals, i.e., positive and negative atoms.\<close>
datatype 'a literal = PLit 'a | NLit 'a

text \<open>
  A clause is a list of literals (interpreted as
  disjunction of literals).
\<close>

type_synonym 'a clause = "('a literal) list"

text \<open>
  A conjunctive normal form (CNF) is a list of clauses
  (interpreted as conjunction of clauses).
\<close>

type_synonym 'a cnf = "('a clause) list"

abbreviation "Top \<equiv> Neg Bot"

fun of_literal :: "'a literal \<Rightarrow> 'a form"
where
  "of_literal (PLit x) = Atom x"
| "of_literal (NLit x) = Neg (Atom x)"

fun of_clause :: "'a clause \<Rightarrow> 'a form"
where
  "of_clause [] = Bot"
| "of_clause (x#xs) = Disj (of_literal x) (of_clause xs)"

fun of_cnf :: "'a cnf \<Rightarrow> 'a form"
where
  "of_cnf [] = Top"
| "of_cnf (x#xs) = Conj (of_clause x) (of_cnf xs)"

lemma eval_of_cnf_append [simp]:
  "eval v (of_cnf (xs @ ys)) \<longleftrightarrow> eval v (of_cnf xs) \<and> eval v (of_cnf ys)"
by (induct xs) auto

lemma eval_of_clause_append [simp]:
  "eval v (of_clause (xs @ ys)) \<longleftrightarrow> eval v (of_clause xs) \<or> eval v (of_clause ys)"
by (induct xs) auto

text \<open>Eliminate implications.\<close>
fun impl_free :: "'a form \<Rightarrow> 'a form"
where
  "impl_free Bot = Bot"
| "impl_free (Atom x) = Atom x"
| "impl_free (Neg \<phi>) = Neg (impl_free \<phi>)"
| "impl_free (Conj \<phi> \<psi>) = Conj (impl_free \<phi>) (impl_free \<psi>)"
| "impl_free (Disj \<phi> \<psi>) = Disj (impl_free \<phi>) (impl_free \<psi>)"
| "impl_free (Impl \<phi> \<psi>) = Disj (Neg (impl_free \<phi>)) (impl_free \<psi>)"

text \<open>Eliminating implications does not change the meaning of a formula.\<close>
lemma eval_impl_free [simp]:
  "eval v (impl_free \<phi>) = eval v \<phi>"
by (induct \<phi>) auto

text \<open>Check whether a formula is free of implications.\<close>
fun is_impl_free :: "'a form \<Rightarrow> bool"
where
  "is_impl_free Bot \<longleftrightarrow> True"
| "is_impl_free (Atom x) \<longleftrightarrow> True"
| "is_impl_free (Neg \<phi>) \<longleftrightarrow> is_impl_free \<phi>"
| "is_impl_free (Conj \<phi> \<psi>) \<longleftrightarrow> is_impl_free \<phi> \<and> is_impl_free \<psi>"
| "is_impl_free (Disj \<phi> \<psi>) \<longleftrightarrow> is_impl_free \<phi> \<and> is_impl_free \<psi>"
| "is_impl_free (Impl \<phi> \<psi>) \<longleftrightarrow> False"

lemma is_impl_free_impl_free [simp]:
  "is_impl_free (impl_free \<phi>)"
by (induct \<phi>) auto

text \<open>Push negations towards atoms; also called "Negation Normal Form".\<close>
fun nnf :: "'a form \<Rightarrow> 'a form"
where
  "nnf Bot = Bot"
| "nnf Top = Top"
| "nnf (Atom x) = Atom x"
| "nnf (Neg (Atom x)) = Neg (Atom x)"
| "nnf (Neg (Neg \<phi>)) = nnf \<phi>"
| "nnf (Conj \<phi> \<psi>) = Conj (nnf \<phi>) (nnf \<psi>)"
| "nnf (Disj \<phi> \<psi>) = Disj (nnf \<phi>) (nnf \<psi>)"
| "nnf (Neg (Conj \<phi> \<psi>)) = Disj (nnf (Neg \<phi>)) (nnf (Neg \<psi>))"
| "nnf (Neg (Disj \<phi> \<psi>)) = Conj (nnf (Neg \<phi>)) (nnf (Neg \<psi>))"
| "nnf _ = undefined"

text \<open>
  We use a special version of structural induction arising from
  the function definition of @{const nnf}: induction along the
  computation of @{const nnf}. This works since @{const nnf}
  terminates.
\<close>
lemma eval_nnf [simp]:
  "is_impl_free \<phi> \<longrightarrow> eval v (nnf \<phi>) = eval v \<phi>"
by (induct \<phi> rule: nnf.induct) auto

fun is_nnf :: "'a form \<Rightarrow> bool"
where
  "is_nnf Bot \<longleftrightarrow> True"
| "is_nnf Top \<longleftrightarrow> True"
| "is_nnf (Atom x) \<longleftrightarrow> True"
| "is_nnf (Neg (Atom x)) \<longleftrightarrow> True"
| "is_nnf (Conj \<phi> \<psi>) \<longleftrightarrow> is_nnf \<phi> \<and> is_nnf \<psi>"
| "is_nnf (Disj \<phi> \<psi>) \<longleftrightarrow> is_nnf \<phi> \<and> is_nnf \<psi>"
| "is_nnf _ \<longleftrightarrow> False"

lemma is_nnf_nnf [simp]:
  "is_impl_free \<phi> \<longrightarrow> is_nnf (nnf \<phi>)"
by (induct \<phi> rule: nnf.induct) auto

text \<open>Distribute disjunctions over conjunctions.\<close>
fun distr :: "'a cnf \<Rightarrow> 'a cnf \<Rightarrow> 'a cnf"
where
  "distr [] ys = []"
| "distr (x#xs) ys = map (\<lambda>y. x @ y) ys @ distr xs ys"

lemma eval_of_cnf_map [simp]:
  "eval v (of_cnf (map (\<lambda>y. x @ y) ys)) \<longleftrightarrow> eval v (of_clause x) \<or> eval v (of_cnf ys)"
by (induct ys) auto

lemma eval_of_cnf_distr [simp]:
  "eval v (of_cnf (distr xs ys)) \<longleftrightarrow> eval v (of_cnf xs) \<or> eval v (of_cnf ys)"
by (induct xs) auto

fun cnf :: "'a form \<Rightarrow> 'a cnf"
where
  "cnf Bot = [[]]"
| "cnf Top = []"
| "cnf (Atom x) = [[PLit x]]"
| "cnf (Neg (Atom x)) = [[NLit x]]"
| "cnf (Conj \<phi> \<psi>) = cnf \<phi> @ cnf \<psi>"
| "cnf (Disj \<phi> \<psi>) = distr (cnf \<phi>) (cnf \<psi>)"
| "cnf _ = undefined"

lemma eval_of_cnf_cnf [simp]:
  "is_nnf \<phi> \<longrightarrow> eval v (of_cnf (cnf \<phi>)) = eval v \<phi>"
by (induct \<phi> rule: cnf.induct) auto

lemma eval_cnf [simp]:
  "eval v (of_cnf (cnf (nnf (impl_free \<phi>)))) = eval v \<phi>"
by (auto)


section \<open>Tseitin's Transformation\<close>

fun tseitin :: "'a form \<Rightarrow> 'a form cnf"
where
 "tseitin Bot = [[NLit Bot]]"
|"tseitin (Atom x) = []"
| "tseitin (Neg \<phi>) = [PLit (Neg \<phi>), PLit \<phi>] # [NLit (Neg \<phi>), NLit \<phi>] # tseitin \<phi>"
| "tseitin (Conj \<phi> \<psi>) =
    [NLit (Conj \<phi> \<psi>), PLit \<phi>] # [NLit (Conj \<phi> \<psi>), PLit \<psi>] # [PLit (Conj \<phi> \<psi>), NLit \<phi>, NLit \<psi>] #
      tseitin \<phi> @ tseitin \<psi>"
| "tseitin (Disj \<phi> \<psi>) =
    [PLit (Disj \<phi> \<psi>), NLit \<phi>] # [PLit (Disj \<phi> \<psi>), NLit \<psi>] # [NLit (Disj \<phi> \<psi>), PLit \<phi>, PLit \<psi>] #
      tseitin \<phi> @ tseitin \<psi>"
| "tseitin (Impl \<phi> \<psi>) =
    [PLit (Impl \<phi> \<psi>), PLit \<phi>] # [PLit (Impl \<phi> \<psi>), NLit \<psi>] # [NLit (Impl \<phi> \<psi>), NLit \<phi>, PLit \<psi>] #
      tseitin \<phi> @ tseitin \<psi>"

lemma eval_eval_tseitin:
  "eval (eval v) (of_cnf (tseitin \<phi>))"
by (induct \<phi>) auto

lemma eval_tseitin_imp_eval [simp]:
  "eval v (of_cnf (tseitin \<phi>)) \<longrightarrow> eval (v \<circ> Atom) \<phi> = v \<phi>"
by (induct \<phi>) auto

text \<open>Tseitin's transformation produces equisatisfiable formulas.\<close>
lemma sat_tseitin [simp]:
  "sat (of_cnf ([PLit \<phi>] # tseitin \<phi>)) \<longleftrightarrow> sat \<phi>"
text \<open>
  \<bullet> \<^bold>N\<^bold>o\<^bold>t\<^bold>e\<^bold>: the left-hand formula is equivalent to
    @{term "Conj (Atom \<phi>) (of_cnf (tseitin \<phi>))"}
\<close>
proof
  assume "sat (of_cnf ([PLit \<phi>] # tseitin \<phi>))"
  then obtain v where "eval v (of_cnf ([PLit \<phi>] # tseitin \<phi>))" by (auto simp: sat_def)
  then have "eval (v \<circ> Atom) \<phi>" by auto
  then show "sat \<phi>" by (auto simp: sat_def)
next
  assume "sat \<phi>"
  then obtain v where "eval v \<phi>" by (auto simp: sat_def)
  then have "eval (eval v) (of_cnf ([PLit \<phi>] # tseitin \<phi>))" by (auto simp: eval_eval_tseitin)
  then show "sat (of_cnf ([PLit \<phi>] # tseitin \<phi>))" by (auto simp: sat_def)
qed


section \<open>Binary Trees\<close>

datatype 'a btree = Empty | Node 'a "'a btree" "'a btree"

fun flatten :: "'a btree \<Rightarrow> 'a list"
where
  "flatten Empty = []"
| "flatten (Node x l r) = flatten l @ x # flatten r"

fun insert :: "'a::linorder \<Rightarrow> 'a btree \<Rightarrow> 'a btree"
where
  "insert x Empty = Node x Empty Empty"
| "insert x (Node y l r) =
    (if x < y then Node y (insert x l) r
    else Node y l (insert x r))"

lemma set_flatten_insert [simp]:
  "set (flatten (insert x t)) = {x} \<union> set (flatten t)"
by (induct t) auto

lemma sorted_flatten_insert [simp]:
  "sorted (flatten (insert x t)) \<longleftrightarrow> sorted (flatten t)"
by (induct t) (auto simp: sorted_Cons sorted_append)

lemma sorted_flatten_fold_insert [simp]:
  "\<forall>t. sorted (flatten (fold insert xs t)) \<longleftrightarrow> sorted (flatten t)"
by (induct xs) auto

abbreviation "sort' xs \<equiv> flatten (fold insert xs Empty)"

corollary sorted_sort': "sorted (sort' xs)" by simp

fun count :: "'a list \<Rightarrow> 'a \<Rightarrow> nat"
where
  "count [] y = 0"
| "count (x#xs) y = (if x = y then count xs y + 1 else count xs y)"

lemma count_append [simp]:
  "count (xs @ ys) y = count xs y + count ys y"
by (induct xs) auto

lemma count_flatten_insert [simp]:
  "count (flatten (insert x t)) y =
    (if x = y then count (flatten t) y + 1 else count (flatten t) y)"
by (induct t) auto

lemma count_flatten_fold_insert [simp]:
  "\<forall>t. count (flatten (fold insert xs t)) y = count xs y + count (flatten t) y"
by (induct xs) auto

corollary count_sort': "count (sort' xs) y = count xs y" by simp

end
